import os, time
import psutil
from collections import namedtuple
import system_info_aix
import whatap.util.process_util as process_util
import whatap.util.thread_util as thread_util
import whatap.util.logging_util as logging_util
import whatap.agent.conf.configure as conf
from StringIO import StringIO

from . import _DiskPerf

filesystem_blacklist = [
    "procfs"
]
def iterateLines(lines, skip=0):
    for i, l in enumerate(lines.split('\n')):
        l = l.strip()
        if i < skip or not l:
            continue
        
        yield l.split()

_mountPointCache = None
def parseFromMountPoint(mountPoint):
    global _mountPointCache
    if _mountPointCache is None:
        _mountPointCache = {}
        lspv = process_util.executeCmd('lspv')
        
        for l in lspv.split('\n'):
            l = l.strip()
            if not l:
                continue
            (deviceId, serial, vgid, stauts)= l.split()
            lspvlist = process_util.executeCmd('lsvg -l %s'%(vgid))
            for tokens in iterateLines(lspvlist, skip=2):
                if len(tokens) < 7:
                    continue
                lvname, lvtype, lvlps, lvpps, lvpvs, lvstate, mountpoint = tokens
                _mountPointCache[mountpoint] = deviceId
    return _mountPointCache.get(mountPoint)


_mountPointCacheEx = None
def parseFromMountPointEx(mountPoint):
    global _mountPointCacheEx
    if _mountPointCacheEx is None:
        _mountPointCacheEx = {}
        lspv = process_util.executeCmd('lspv')
        
        for l in lspv.split('\n'):
            l = l.strip()
            if not l or len(l.split()) < 4:
                continue
            (deviceId, serial, vgid, stauts)= l.split()
            if 'none' == vgid.lower():
                continue
            lspvlist = process_util.executeCmd('lsvg -l %s'%(vgid))
            for tokens in iterateLines(lspvlist, skip=2):
                if len(tokens) < 7:
                    continue
                lvname, lvtype, lvlps, lvpps, lvpvs, lvstate, mountpoint = tokens
                if 'N/A' == mountpoint:
                    continue
                if mountpoint not in _mountPointCacheEx:
                    _mountPointCacheEx[mountpoint] = []    
                _mountPointCacheEx[mountpoint].append(deviceId)    
    return _mountPointCacheEx.get(mountPoint)

sdiskio = namedtuple('sdiskio', ['read_count', 'write_count',
                                 'read_bytes', 'write_bytes',
                                 'read_time', 'write_time', 'time'])

class _DiskIoPerfCache:
    def __init__(self, iocounters=None):
        self.iocounters= iocounters
        self.timestamp = time.time()

    def contains(self, mountPoint):
        deviceIds = parseFromMountPointEx(mountPoint)
        deviceFound = False
        if not deviceIds:
            return deviceFound
        for deviceId in deviceIds:
            deviceFound = deviceFound or deviceId in self.iocounters
        return deviceFound
    
    def get(self, mountPoint):
        deviceIds = parseFromMountPointEx(mountPoint)
        iotime = 0
        read_count = 0
        write_count = 0
        read_bytes = 0
        write_bytes = 0
        read_time = 0
        write_time = 0

        for deviceId in deviceIds:
            iocounter = self.iocounters.get(deviceId)
            iotime += iocounter.time
            read_count += iocounter.read_count
            write_count += iocounter.write_count
            read_bytes += iocounter.read_bytes
            write_bytes += iocounter.write_bytes
            read_time += iocounter.read_time
            write_time += iocounter.write_time
        iotime /= len(deviceIds)
        
        return sdiskio(time=iotime, read_count=read_count, write_count=write_count, read_bytes=read_bytes, write_bytes=write_bytes, read_time=read_time, write_time=write_time)

iocountersLastTime = _DiskIoPerfCache(psutil.disk_io_counters(perdisk=True))

def populate(mountPoint, callback, iocounters, iocountersLastTime):
    
    if not iocounters.contains(mountPoint) or not iocountersLastTime.contains(mountPoint):
        return
    iocounter = iocounters.get(mountPoint)
    iocounterLasttime =  iocountersLastTime.get(mountPoint)
    timediff = float(iocounters.timestamp - iocountersLastTime.timestamp)

    ioutil = float(100) * float(iocounter.time - iocounterLasttime.time) / float(system_info_aix.clk_tck) / timediff
    iopsRead = float(iocounter.read_count - iocounterLasttime.read_count) / timediff
    iopsWrite = float(iocounter.write_count - iocounterLasttime.write_count) / timediff
    bpsRead = float(iocounter.read_bytes - iocounterLasttime.read_bytes) /timediff
    bpsWrite = float(iocounter.write_bytes - iocounterLasttime.write_bytes) /timediff
    callback(ioutil, iopsRead,iopsWrite, bpsRead, bpsWrite )

def populateIoPerf(diskQuotas):
    global iocountersLastTime

    iocounters = _DiskIoPerfCache(psutil.disk_io_counters(perdisk=True))
    diskQuotasThisTime = []
    for diskQuota in diskQuotas:
        diskQuotasThisTime.append(diskQuota)
        mountPoint = diskQuota.mountPoint

        if not iocounters.contains(mountPoint):
            continue
        
        def setDiskIo(ioPercent, readiops, writeiops, readbps, writebps):
            diskQuota.readIops = readiops
            diskQuota.writeIops = writeiops
            diskQuota.readBps = readbps
            diskQuota.writeBps = writebps
            diskQuota.ioPercent = ioPercent


        populate(mountPoint, setDiskIo, iocounters, iocountersLastTime )
    
    iocountersLastTime = iocounters 

    return diskQuotasThisTime

def updateDiskPerf():
    global lastDiskPerf
    diskQuotas = []
    for i, part in enumerate(psutil.disk_partitions(all=False)):
        if part.fstype in filesystem_blacklist:
            continue
        p = _DiskPerf()
        p.mountPoint = part.mountpoint
        if not p.mountPoint:
            p.mountPoint = part.device
        p.deviceId = part.device
        p.fileSystem = part.fstype
        f_bsize, f_frsize, f_blocks, f_bfree, f_bavail, f_files, f_ffree, f_favail, f_flag, f_namemax = os.statvfs(p.mountPoint)
        p.totalSpace = int(f_frsize * f_blocks)
        p.usedSpace = int(f_frsize * (f_blocks - f_bfree))
        p.usedPercent = float(100.0 * float(f_blocks - f_bfree) / float(f_blocks))
        p.freeSpace = p.totalSpace - p.usedSpace
        p.freePercent = float(100.0 * f_bfree / f_blocks)
        p.blksize = int(f_frsize)
        p.inodePercent = float(100.0 * float(f_files - f_ffree) / float(f_files))

        diskQuotas.append(p)
        
    lastDiskPerf= populateIoPerf(diskQuotas)

lastDiskPerf = None
def getDisk():
    global lastDiskPerf
    return lastDiskPerf

try:
    updateDiskPerf()    
    thread_util.asyncLoop(updateDiskPerf, conf.GetConfig().DiskstatInterval)
except Exception, e:
    logging_util.debug("disk perf aix error:", e)


oldIOTotal = None
def GetDiskTotalIO():
    disk = psutil.disk_io_counters()
    now = time.time()

    global oldIOTotal
    if not oldIOTotal:
        oldIOTotal = [disk, now]
        return None

    oldDisk = oldIOTotal[0]
    diffTime = float(now - oldIOTotal[1])

    ret = _DiskPerf()

    ret.readBps = float(disk.read_bytes - oldDisk.read_bytes) / diffTime
    ret.writeBps = float(disk.write_bytes - oldDisk.write_bytes) / diffTime
    ret.readIops = float(disk.read_count - oldDisk.read_count) / diffTime
    ret.writeIops = float(disk.write_count - oldDisk.write_count) / diffTime

    oldIOTotal = [disk, now]
    return ret

oldIO = None
def GetDiskIO():
    disk = psutil.disk_io_counters(perdisk=True)
    now = time.time()

    global oldIO
    if not oldIO:
        oldIO = [disk, now]
        return None

    oldDisk = oldIO[0]
    diffTime = float(now - oldIO[1])

    retArr = []
    for k, v in disk.items():
        if k not in oldDisk or k.startswith("cd"):
            continue
        oldv = oldDisk[k]

        ret = _DiskPerf()
        ret.deviceId = k
        ret.readBps = float(v.read_bytes - oldv.read_bytes) / diffTime
        ret.writeBps = float(v.write_bytes - oldv.write_bytes) / diffTime
        readIoCount  = float(v.read_count - oldv.read_count)
        ret.readIops = readIoCount / diffTime
        writeIoCount = float(v.write_count - oldv.write_count)
        ret.writeIops = writeIoCount / diffTime

        readTime = float(v.read_time - oldv.read_time)
        writeTime = float(v.write_time - oldv.write_time)

        busyTime = float(v.time - oldv.time)
        if readIoCount == 0:
            ret.readTime = 0
        else:
            ret.readTime = readTime / readIoCount

        if writeIoCount == 0:
            ret.writeTime = 0
        else:
            ret.writeTime = writeTime / writeIoCount
        
        ret.ioPercent = 100 * (busyTime) / (diffTime * 1000)
        if ret.ioPercent > float(100):
            ret.ioPercent = float(100)

        retArr.append(ret)

    oldIO = [disk, now]
    return retArr

        
        

def test():
    for a in getDisk():
        print a

if __name__ == '__main__':
    test()
